
package asit.moa.client.util;

import iaik.asn1.structures.Name;
import iaik.ixsil.algorithms.CanonicalizationAlgorithmImplCanonicalXML;
import iaik.ixsil.algorithms.DigestAlgorithmImplSHA1;
import iaik.ixsil.algorithms.SignatureAlgorithmImplRSA;
import iaik.ixsil.algorithms.Transform;
import iaik.ixsil.core.Position;
import iaik.ixsil.core.Signer;
import iaik.ixsil.core.SignerReference;
import iaik.ixsil.core.SignerSignature;
import iaik.ixsil.core.SignerSignedInfo;
import iaik.ixsil.exceptions.DOMUtilsException;
import iaik.ixsil.init.IXSILConstants;
import iaik.ixsil.keyinfo.KeyManagerImpl;
import iaik.ixsil.keyinfo.x509.KeyProviderImplX509Data;
import iaik.ixsil.keyinfo.x509.X509Data;
import iaik.ixsil.util.URI;
import iaik.security.rsa.RSAPrivateKey;
import iaik.utils.RFC2253NameParserException;
import iaik.utils.Util;
import iaik.x509.X509Certificate;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.SAXException;
import asit.common.Utils;
import asit.moa.client.exception.ClientException;


/**
 * @author atauber
 */
public class MoaUtils {

  public final static String ETSI_NS = "http://uri.etsi.org/01903/v1.1.1#";
  public final static String ETSI_PREFIX = "etsi";
  public final static String DSIG_PREFIX = "dsig";
  public final static String SHA1_URI = "http://www.w3.org/2000/09/xmldsig#sha1";
  public final static String DSIG_NS = "http://www.w3.org/2000/09/xmldsig##";
  
  public static Element createXAeDS(String sigTime, String signatureId, String referenceId,
      X509Certificate certificate, Document doc) throws DOMException, RFC2253NameParserException,
      DOMUtilsException, ParserConfigurationException, SAXException, IOException {
    // QualifyingProperties
    Element qualifyingProperties = doc.createElement(ETSI_PREFIX
                                                                + ":QualifyingProperties");
    qualifyingProperties.setAttribute("xmlns:dsig", DSIG_NS);
    qualifyingProperties.setAttribute("Target", "#" + signatureId);
    qualifyingProperties.setAttribute("xmlns:" + ETSI_PREFIX, ETSI_NS);
    // SignedProperties
    Element signedProperties = doc.createElement(ETSI_PREFIX + ":SignedProperties");
    qualifyingProperties.appendChild(signedProperties);
    // SignedSignatureProperties
    Element signedSignatureProperties = doc
        .createElement(ETSI_PREFIX + ":SignedSignatureProperties");
    signedProperties.appendChild(signedSignatureProperties);
    // SigningTime
    Element signingTime = doc.createElement(ETSI_PREFIX + ":SigningTime");
    signingTime.appendChild(doc.createTextNode(sigTime));
    signedSignatureProperties.appendChild(signingTime);
    // SigningCertificate
    Element signingCertificate = doc.createElement(ETSI_PREFIX + ":SigningCertificate");
    signedSignatureProperties.appendChild(signingCertificate);
    // Cert
    Element cert = doc.createElement(ETSI_PREFIX + ":Cert");
    signingCertificate.appendChild(cert);
    // CertDigest
    Element certDigest = doc.createElement(ETSI_PREFIX + ":CertDigest");
    cert.appendChild(certDigest);
    // DigestMethod
    Element digestMethod = doc.createElement(ETSI_PREFIX + ":DigestMethod");
    digestMethod.setAttribute("Algorithm", SHA1_URI);
    certDigest.appendChild(digestMethod);
    // DigestValue
    Element digestValue = doc.createElement(ETSI_PREFIX + ":DigestValue");
    digestValue.appendChild(doc.createTextNode(new String(Util.Base64Encode(certificate
        .getFingerprintSHA()))));
    certDigest.appendChild(digestValue);
    // IssuerSerial
    Element issuerSerial = doc.createElement(ETSI_PREFIX + ":IssuerSerial");
    cert.appendChild(issuerSerial);
    // X509IssuerName
    Element x509IssuerName = doc.createElement(DSIG_PREFIX + ":X509IssuerName");
    Name issuer = (Name) certificate.getIssuerDN();
    x509IssuerName.appendChild(doc.createTextNode(issuer.getRFC2253String()));
    issuerSerial.appendChild(x509IssuerName);
    // X509SerialNumber
    Element x509SerialNumber = doc.createElement(DSIG_PREFIX + ":X509SerialNumber");
    x509SerialNumber.appendChild(doc.createTextNode(certificate.getSerialNumber().toString()));
    issuerSerial.appendChild(x509SerialNumber);
    // SignaturePolicyIdentifyer
    Element signaturePolicyIdentifier = doc
        .createElement(ETSI_PREFIX + ":SignaturePolicyIdentifier");
    signedSignatureProperties.appendChild(signaturePolicyIdentifier);
    // SignaturePolicyImplied
    Element signaturePolicyImplied = doc.createElement(ETSI_PREFIX
                                                                  + ":SignaturePolicyImplied");
    signaturePolicyIdentifier.appendChild(signaturePolicyImplied);
    // SignatureDataObjectProperties
    Element signedDataObjectProperties = doc.createElement(ETSI_PREFIX + ":SignedDataObjectProperties");
    signedProperties.appendChild(signedDataObjectProperties);
    // DataObjectFormat
    Element dataObjectFormat = doc.createElement(ETSI_PREFIX + ":DataObjectFormat");
    dataObjectFormat.setAttribute("ObjectReference", "#" + referenceId);
    signedDataObjectProperties.appendChild(dataObjectFormat);
    // MimeType
    Element mimeType = doc.createElement(ETSI_PREFIX + ":MimeType");
    mimeType.appendChild(doc.createTextNode("text/html"));
    dataObjectFormat.appendChild(mimeType);
    ByteArrayOutputStream out = new ByteArrayOutputStream();
    Utils.serializeElement(qualifyingProperties, out);
    DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
    factory.setNamespaceAware(true);
    DocumentBuilder builder;
    builder = factory.newDocumentBuilder();
    Element ret = builder.parse(new ByteArrayInputStream(out.toByteArray())).getDocumentElement();
    ret.removeAttribute("xmlns:dsig");
    return ret;
    //return qualifyingProperties;
  }

  public static Document generateNotificationSignature(Document notification, String signatureName,
      String referenceName, String etsiId, String signingTime, Transform[] transforms,
      String xPathPosition, int position, File pK, File verifyCertificate) throws ClientException {
    try {
      Position sigPosition = new Position(xPathPosition, "b http://reference.e-government.gv.at/namespace/notificationdata/20040510#", position);
      Signer signer = new Signer(notification, new URI(IXSILConstants.DUMMY_ABSURI_), sigPosition); //notification,
      // new
      // URI(IXSILConstants.DUMMY_ABSURI_),
      // sigPosition);
      // create SignedInfo
      SignerSignature signature = signer.getSignature();
      SignerSignedInfo signedInfo = signature.getSignerSignedInfo();
      // canonicalization algorithm
      CanonicalizationAlgorithmImplCanonicalXML c14nAlg = new CanonicalizationAlgorithmImplCanonicalXML();
      signedInfo.setCanonicalizationAlgorithm(c14nAlg);
      // Set data for creating signature
      SignatureAlgorithmImplRSA signatureAlg = new SignatureAlgorithmImplRSA();
      RSAPrivateKey privateKey = loadPrivateKey(pK);
      signatureAlg.setSignerKey(privateKey);
      signedInfo.setSignatureAlgorithm(signatureAlg);
      // Create reference
      SignerReference reference = signedInfo.createReference();
      reference.setId(referenceName);
      reference.setURI(new URI(""));
      DigestAlgorithmImplSHA1 digestAlg = new DigestAlgorithmImplSHA1();
      reference.setDigestAlgorithm(digestAlg);
      // set transformations
      if (transforms != null) {
        for (int i = 0; i < transforms.length; i++) {
          reference.insertTransformAt(transforms[i], i);
        }
      }
      signedInfo.addReference(reference);
      // add signers certificate
      X509Certificate cert = new X509Certificate(new FileInputStream(verifyCertificate));
      X509Data x509Data = new X509Data();
      x509Data.insertHintAt(cert, 0);
      KeyManagerImpl keyManager = new KeyManagerImpl(signer.toDocument());
      KeyProviderImplX509Data x509DataKeyProvider = new KeyProviderImplX509Data(signer.toDocument());
      x509DataKeyProvider.insertX509DataAt(x509Data, 0);
      keyManager.addKeyProvider(x509DataKeyProvider);
      signer.getSignature().setKeyManager(keyManager);
      // Add etsi-QualifyingProperties (signed additional info)
      Element etsi = createXAeDS(signingTime, signatureName, referenceName, cert, signer
          .toDocument());
      iaik.ixsil.core.Object object = signature.createObject(etsi);
      object.setId(etsiId);
      signer.getSignature().addObject(object);
      SignerReference etsiRef = signedInfo.createReference();
      etsiRef.setType(new URI("http://uri.etsi.org/01903/v1.1.1#SignedProperties"));
      etsiRef
          .setURI(new URI("#xmlns(etsi=http://uri.etsi.org/01903/v1.1.1%23)%20xpointer(id('"
                          + etsiId
                          + "')/child::etsi:QualifyingProperties/child::etsi:SignedProperties)"));
      etsiRef.setDigestAlgorithm(digestAlg);
      signedInfo.addReference(etsiRef);
      // End etsi-QualifyingProperties
      signer.getSignature().sign();
      Document signedDoc = signer.toDocument();
      /*
       * ByteArrayOutputStream out = new ByteArrayOutputStream();
       * DOMUtilsImpl.serializeDocument(signedDoc, out); // re-parse for
       * signature value replacing DocumentBuilderFactory factory =
       * DocumentBuilderFactory.newInstance();
       * //factory.setNamespaceAware(true); DocumentBuilder builder =
       * factory.newDocumentBuilder(); Document doc = builder.parse(new
       * ByteArrayInputStream(out.toByteArray())); NodeList signatures =
       * XPathAPI.selectNodeList(doc, xPathPosition +
       * "/Signature/SignatureValue/text()"); // exchange signature value if
       * (signatures.getLength() > 0) {
       * signatures.item(0).setNodeValue(signatureValue); }
       */
      return signedDoc;
    } catch (Exception e) {
      e.printStackTrace();
      throw new ClientException(e);
    }
  }

  private static RSAPrivateKey loadPrivateKey(File privateKey) throws Exception {
    FileInputStream rSAPrivateKeyFIS = new FileInputStream(privateKey);
    return new RSAPrivateKey(rSAPrivateKeyFIS);
  }
}