package asit.moa.client.verify;

import iaik.asn1.structures.Name;
import iaik.ixsil.util.Utils;
import iaik.utils.RFC2253NameParser;
import iaik.x509.X509Certificate;

import java.math.BigInteger;
import java.util.Date;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.xpath.XPathAPI;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import asit.moa.client.exception.ClientException;

/**
 * @author <a href="mailto:arne.tauber@iaik.tugraz.at">Arne Tauber</a>
 * @author modified by <a href="mailto:thomas.knall@iaik.tugraz.at">Thomas Knall</a>
 */
class VerifyResultImpl implements VerifyResult {

	protected static Log log_ = LogFactory.getLog(VerifyResultImpl.class.getName());
	protected Name subjectName_ = null;
	protected Name issuerName_ = null;
	protected BigInteger serial_ = null;
	protected X509Certificate cert_ = null;
	protected boolean isAuthority_ = false;
	protected boolean isQualified_ = false;
	protected String authorityCode_ = null;
	protected int sigCheckCode_ = -1;
	protected int manifestCheckCode_ = -1;
	protected int xmldsigManifestCheckCode_ = -1;
	protected int certCheckCode_ = -1;
	protected int[] failedSignatureCheckReferences_ = null;
	protected int[] failedManifestCheckReferences_ = null;
	protected Element verifyXMLSignatureResponse_ = null;
	protected Date dateOfVerification_ = null;
	protected HashInputData[] hashInputDataElements = null;

	public VerifyResultImpl(Element element) throws ClientException {
		if (element == null) {
			throw new NullPointerException("Element must not be null.");
		}

		this.hashInputDataElements = null;
		this.verifyXMLSignatureResponse_ = element;

		String moa = element.getPrefix();
		if (moa == null) {
			moa = "";
		}
		
		NodeList nodeList = element.getElementsByTagNameNS("http://www.w3.org/2000/09/xmldsig#", "X509Data");
		String dsig = "";
		if (nodeList != null && nodeList.getLength() > 0) {
			dsig = nodeList.item(0).getPrefix();
		}
		if (dsig == null) {
			dsig = "";
		}

	   String rootSelect     = "self::" + moa + ":VerifyXMLSignatureResponse/";
		String x509DataSelect = rootSelect + moa + ":SignerInfo/" + dsig + ":X509Data";

		try {
			
			// X509 DATA NODE
			Node x509DataNode = XPathAPI.selectSingleNode(element, x509DataSelect);
			if (x509DataNode != null) {
				
				// SUBJECT NAME
				Node subjectName = XPathAPI.selectSingleNode(x509DataNode, dsig + ":X509SubjectName/text()");
				if (subjectName != null) {
					log_.debug("Found X509 Subject DN: " + subjectName.getNodeValue());
					RFC2253NameParser parser = new RFC2253NameParser(subjectName.getNodeValue());
					this.subjectName_ = parser.parse();
				}
				
				// X509 ISSUER/SERIAL NODE
				Node x509IssuerSerialNode = XPathAPI.selectSingleNode(x509DataNode, dsig + ":X509IssuerSerial");
				if (x509IssuerSerialNode != null) {
					
					// ISSUER NAME
					Node issuerName = XPathAPI.selectSingleNode(x509IssuerSerialNode, dsig + ":X509IssuerName/text()");
					if (issuerName != null) {
						log_.debug("Found X509 Issuer DN: " + issuerName.getNodeValue());
						RFC2253NameParser parser = new RFC2253NameParser(issuerName.getNodeValue());
						this.issuerName_ = parser.parse();
					}

					// ISSUER SERIAL
					// note: its the subject's certificate serial not the issuer's
					Node serial = XPathAPI.selectSingleNode(x509IssuerSerialNode, dsig + ":X509SerialNumber/text()");
					if (serial != null) {
						log_.debug("Found X509 Serial: " + serial.getNodeValue());
						this.serial_ = new BigInteger(serial.getNodeValue());
					}

				}

				// CERTIFICATE
				Node cert = XPathAPI.selectSingleNode(x509DataNode, dsig	+ ":X509Certificate/text()");
				if (cert != null) {
					log_.debug("Found X509 Certficate:");
					this.cert_ = new X509Certificate(Utils.base64Decode(cert.getNodeValue().getBytes()));
				}

				// PUBLIC AUTHORITY
				Node publicAuthority = XPathAPI.selectSingleNode(x509DataNode, "./" + moa + ":PublicAuthority");
				if (publicAuthority != null) {
					log_.debug("Found Public Authority");
					this.isAuthority_ = true;
					Node authorityCode = XPathAPI.selectSingleNode(publicAuthority, "./" + moa + ":Code/text()");
					if (authorityCode != null) {
						log_.debug("Found authority code: " + authorityCode.getNodeValue());
						this.authorityCode_ = authorityCode.getNodeValue();
					}
				}

				// QUALIFIED CERTIFICATE
				Node qualifiedCertificate = XPathAPI.selectSingleNode(x509DataNode, "./" + moa + ":QualifiedCertificate");
				if (qualifiedCertificate != null) {
					log_.debug("Found Qualified Certificate");
					this.isQualified_ = true;
				}

			}


			// SIGNATURE CHECK CODE
			Node sigCheckNode = XPathAPI.selectSingleNode(element, rootSelect + moa	+ ":SignatureCheck/" + moa + ":Code/text()");
			if (sigCheckNode != null) {
				log_.debug("Found Signature Check code: " + sigCheckNode.getNodeValue());
				this.sigCheckCode_ = Integer.parseInt(sigCheckNode.getNodeValue());
			}

			// FAILED SIGNATURE REFERENCES
			NodeList failedSigRefs = XPathAPI.selectNodeList(element, rootSelect + moa	+ ":SignatureCheck/" + moa + ":Info/" + moa + ":FailedReference/text()");
			log_.debug("Found " + failedSigRefs.getLength() + " failed signature references");
			this.failedSignatureCheckReferences_ = new int[failedSigRefs.getLength()];
			for (int i = 0; i < failedSigRefs.getLength(); i++) {
				this.failedSignatureCheckReferences_[i] = Integer.parseInt(failedSigRefs.item(i).getNodeValue());
			}

			
			// MANIFEST CHECK CODE
			Node manifestCheckNode = XPathAPI.selectSingleNode(element, rootSelect + moa + ":SignatureManifestCheck/" + moa + ":Code/text()");
			if (manifestCheckNode != null) {
				log_.debug("Found Manifest Check code: " + manifestCheckNode.getNodeValue());
				this.manifestCheckCode_ = Integer.parseInt(manifestCheckNode.getNodeValue());
			}

			// FAILED MANIFEST REFERENCES
			NodeList failedManifestRefs = XPathAPI.selectNodeList(element, rootSelect + moa + ":SignatureManifestCheck/" + moa + ":Info/" + moa + ":FailedReference/text()");
			log_.debug("Found " + failedManifestRefs.getLength() + " failed manifest references");
			this.failedManifestCheckReferences_ = new int[failedManifestRefs.getLength()];
			for (int i = 0; i < failedManifestRefs.getLength(); i++) {
				this.failedManifestCheckReferences_[i] = Integer.parseInt(failedManifestRefs.item(i).getNodeValue());
			}


         // XMLDSIG CHECK CODE
         Node xmldsigManifestCheckNode = XPathAPI.selectSingleNode(element, rootSelect + moa + ":XMLDSIGManifestCheck/" + moa + ":Code/text()");
         if (xmldsigManifestCheckNode != null) {
            log_.debug("Found XMLDSIG Manifest Check code: " + xmldsigManifestCheckNode.getNodeValue());
            this.xmldsigManifestCheckCode_ = Integer.parseInt(xmldsigManifestCheckNode.getNodeValue());
         }

         
         // CERTIFICATE CHECK CODE
			Node certCheckNode = XPathAPI.selectSingleNode(element, rootSelect + moa + ":CertificateCheck/" + moa + ":Code/text()");
			if (certCheckNode != null) {
				log_.debug("Found Certificate Check code: " + certCheckNode.getNodeValue());
				this.certCheckCode_ = Integer.parseInt(certCheckNode.getNodeValue());
//				this.certCheckCode_ = 0;
//				log_.warn("Modified version of Response Parser !!! (getCertificateCheckCode always returns true)");
			}
			
			// HASH INPUT DATA
			NodeList hashInputDataNodes = XPathAPI.selectNodeList(element, rootSelect + moa + ":HashInputData");
         this.hashInputDataElements = null;
			if (hashInputDataNodes != null) {
			   log_.debug("Found HashInputData node(s).");
			   this.hashInputDataElements = new HashInputData[hashInputDataNodes.getLength()];
			   for (int i = 0; i < hashInputDataNodes.getLength(); i++) {
	            HashInputData hashInputData = new HashInputData();
			      Node hashInputDataNode = hashInputDataNodes.item(i);
			      if (Node.ELEMENT_NODE == hashInputDataNode.getNodeType()) {
			         Element hashInputDataElement = (Element) hashInputDataNode;
			         hashInputData.setPartOf(hashInputDataElement.getAttribute("PartOf"));
			         String rsr = hashInputDataElement.getAttribute("ReferringSigReference");
			         if (rsr != null && rsr.length() > 0) {
			            hashInputData.setReferringSigReference(new Integer(rsr));
			         }
			      }
               Node base64ContentNode = XPathAPI.selectSingleNode(hashInputDataNode,  "./" + moa + ":Base64Content/text()");
               if (base64ContentNode != null) {
                  String base64String = base64ContentNode.getNodeValue();
                  hashInputData.setDecodedContent(Utils.base64Decode(base64String.getBytes()));
               }
               this.hashInputDataElements[i] = hashInputData;
			   }
			}

		} catch (Exception e) {
			throw new ClientException(e);
		}

	}

	/**
	 * @see asit.moa.client.verify.VerifyResult#getSubjectName()
	 */
	public Name getSubjectName() {
		return this.subjectName_;
	}

	/**
	 * @see asit.moa.client.verify.VerifyResult#getIssuerName()
	 */
	public Name getIssuerName() {
		return this.issuerName_;
	}

	/**
	 * Returns the serial number from the {@code X509IssuerSerial}-node. In fact
	 * that node contains the subject's serial number not the issuer's. So ignore
	 * the misleading method name.
	 * 
	 * @return The subject's serial number.
	 * @deprecated Use {@link #getSubjectSerial()} instead.
	 * @see asit.moa.client.verify.VerifyResult#getIssuerSerial()
	 */
	public BigInteger getIssuerSerial() {
		return this.getSubjectSerial();
	}

	/**
	 * Returns the (subject's) serial number (from {@code X509IssuerSerial}-node).
	 * 
	 * @return The subject's serial number.
	 * @see asit.moa.client.verify.VerifyResult#getSubjectSerial()
	 */
	public BigInteger getSubjectSerial() {
		return this.serial_;
	}

	/**
	 * @see asit.moa.client.verify.VerifyResult#getCertificate()
	 */
	public X509Certificate getCertificate() {
		return this.cert_;
	}

	/**
	 * @see asit.moa.client.verify.VerifyResult#isQualifiedCertificate()
	 */
	public boolean isQualifiedCertificate() {
		return this.isQualified_;
	}

	/**
	 * @see asit.moa.client.verify.VerifyResult#isPublicAuthority()
	 */
	public boolean isPublicAuthority() {
		return this.isAuthority_;
	}

	/**
    * @see asit.moa.client.verify.VerifyResult#getPublicAuthorithyCode()
    * @deprecated Use {@link #getPublicAuthorityCode()} instead.
    */
   public String getPublicAuthorithyCode() {
      return getPublicAuthorityCode();
   }

   /**
	 * @see asit.moa.client.verify.VerifyResult#getPublicAuthorityCode()
	 */
	public String getPublicAuthorityCode() {
		return this.authorityCode_;
	}

	/**
	 * @see asit.moa.client.verify.VerifyResult#getSignatureCheckCode()
	 */
	public int getSignatureCheckCode() {
		return this.sigCheckCode_;
	}

	/**
	 * Returns the failed signature references.
	 * 
	 * @return An integer array of failed signature references.
	 * @deprecated Use {@link #getFailedSignatureReferences()} instead.
	 * @see asit.moa.client.verify.VerifyResult#getFailedReferences()
	 */
	public int[] getFailedReferences() {
		return this.getFailedSignatureReferences();
	}

	/**
	 * @see asit.moa.client.verify.VerifyResult#getFailedSignatureReferences()
	 */
	public int[] getFailedSignatureReferences() {
		return this.failedSignatureCheckReferences_;
	}

	/**
	 * @see asit.moa.client.verify.VerifyResult#getCertificateCheckCode()
	 */
	public int getCertificateCheckCode() {
		return this.certCheckCode_;
	}

	/**
	 * @see asit.moa.client.verify.VerifyResult#getFailedManifestReferences()
	 */
	public int[] getFailedManifestReferences() {
		return this.failedManifestCheckReferences_;
	}

	/**
	 * @see asit.moa.client.verify.VerifyResult#getManifestCheckCode()
	 */
	public int getManifestCheckCode() {
		return this.manifestCheckCode_;
	}

   /**
    * @see asit.moa.client.verify.VerifyResult#getXmldsigManifestCheckCode()
    */
   public int getXmldsigManifestCheckCode() {
      return this.xmldsigManifestCheckCode_;
   }

   /**
	 * Returns the MOA {@code VerifyXMLSignatureResponse}.
	 * 
	 * @return The MOA {@code VerifyXMLSignatureResponse}.
	 * @see asit.moa.client.verify.VerifyResult#getVerifyXMLSignatureResponse
	 */
	public Element getVerifyXMLSignatureResponse() {
		return this.verifyXMLSignatureResponse_;
	}

	private static final String toString(int[] a) {
		if (a == null)
			return "null";
		if (a.length == 0)
			return "[]";

		StringBuffer buf = new StringBuffer();
		buf.append('[');
		buf.append(a[0]);

		for (int i = 1; i < a.length; i++) {
			buf.append(", ");
			buf.append(a[i]);
		}

		buf.append("]");
		return buf.toString();
	}

	public String toString() {
		StringBuffer buffer = new StringBuffer();
		buffer.append("subject name = \"").append(this.subjectName_).append("\"");
		buffer.append(", serial number = ").append(this.serial_);
		buffer.append(", issuer name = \"").append(this.issuerName_).append("\"");
		;
		buffer.append(", X509Certificate = ").append(this.cert_ != null ? "available" : "not available");
		buffer.append(", authority = ").append(this.isAuthority_ ? "yes" : "no");
		if (this.authorityCode_ != null) {
			buffer.append(", authority code = ").append(this.authorityCode_);
		}
		buffer.append(", qualified = ").append(this.isQualified_ ? "yes" : "no");
		buffer.append(", signature check code = ").append(this.sigCheckCode_);
		if (this.failedSignatureCheckReferences_ != null) {
			buffer.append(", failed signature check references = ").append(toString(this.failedSignatureCheckReferences_));
		}
      buffer.append(", xmldsigmanifest check code = ").append(this.xmldsigManifestCheckCode_);
      buffer.append(", manifest check code = ").append(this.manifestCheckCode_);
		if (this.failedManifestCheckReferences_ != null) {
			buffer.append(", failed manifest check references = ").append(toString(this.failedManifestCheckReferences_));
		}
		buffer.append(", certificate check code = ").append(this.certCheckCode_);
		return buffer.toString();
	}

	public Date getDateOfVerification() {
		return this.dateOfVerification_;
	}

   /**
    * Returns an array of {@link HashInputData} elements each containing the referenced content of
    * a {@code dsig:Reference} or {@code null} if there are no {@code HashInputData} elements.
    * 
    * @return An array of {@link HashInputData} elements or {@code null}.
    */
	public HashInputData[] getHashInputData() {
	   return this.hashInputDataElements;
   }

}
